#!/bin/bash
# snat_setup.sh
# Script to setup SUSEfirewall2 for SNAT
# Configuration via user_data
# (c) Kurt Garloff <t-systems@garloff.de>, 1/2017
# License: CC-BY-SA 2.0
#
# Syntax: YAML
# otc:
#    snat:
#      masqnet:
#         - NET1
#         - NET2
#      nomasqnet:
#         ...
#      fwdmasq:
#         - NET,DEST,PROTO,SPORT,DPORT
#      ...
#    fail2ban: true
#
#    addip:
#       DEV: IP[/MASK]
#       ...
#
#    addroute:
#       - CIDR via IP dev DEV
#       ...
#    internalnet:
#       - LOCALNET
#     dmznet:
#       - NET1
#       - NET2
#
#    persistroute: true
#
#    movessh: PORT
#
#    SFW2:
#       VAR1: content1
#       VAR2: content2
#
#    autoupdate:
#       frequency: daily|weekly|monthly
#       category: "security recommended optional pkgmanager feature document"
#

# The fields end up changing settings in SUSE /etc/sysconfig/SuSEfirewall2
# See the documentation in that file for the exact meanings
# LOCALNET is translated to the first local network
# INTERNALNET is translated to the networks specified with otc.internalnet:
# DMZNET is translated to the networks specified with dmznet:
# PRIVATENET is translated to 192.168/16 172.16/12 10/8
# PROVIDERNET is translated to 100.64/10
# otc.snat.masqnet -> FW_MASQ_NETS (space separated list)
# otc.snat.nomasqnet -> FW_NOMASQ_NETS (dito)
# otc.snat.fwdmasq -> FW_FORAWRD_MASQ (dito)
# otc.fail2ban -> enable custom script that blocks too many ssh conns
# otc.SFW2.VARx -> overwrite arbitrary SFW2 variables
# otc.autoupdate -> enable unattended package updates
# otc.persistroute -> save routes for reuse on reboot (if you play nasty DHCP tricks)
# otc.movessh -> change ssh to listen on another port (also adjusts fail2ban and SFW2)
#
# Note that LOCALNET is replaced by the local subnet
# The easiest working setup thus looks:
# otc:
#    snat:
#       masqnet:
#          - LOCALNET

if test "$1" == "-x"; then set -x; shift; fi
if test "$1" == "debug"; then DEBUG=1; shift; else unset DEBUG; fi
if test "$1" == "verbose"; then VERBOSE=1; shift; else unset VERBOSE; fi

verbprint()
{
  if test "$VERBOSE" = "1"; then
    echo $@ 1>&2
  fi
}

SFW2CFG_=/etc/sysconfig/SuSEfirewall2
SFW2CUSTCFG_=/etc/sysconfig/scripts/SuSEfirewall2-snathelper
AOUCFG_=/etc/sysconfig/automatic_online_update
SSHDCFG_=/etc/ssh/sshd_config

if test -n "$DEBUG"; then
  cp -p $SFW2CFG_ .
  cp -p $SFW2CUSTCFG_ .
  cp -p $AOUCFG_ .
  sudo cp -p $SSHDCFG_ .
  SFW2CFG=./SuSEfirewall2
  SFW2CUSTCFG=./SuSEfirewall2-snathelper
  AOUCFG=./automatic_online_update
  SSHDCFG=./sshd_config
  sudo chown $UID $SSHDCFG
  get_user_data()
  {
    cat -
  }
else 
  SFW2CFG=$SFW2CFG_
  SFW2CUSTCFG=$SFW2CUSTCFG_
  AOUCFG=$AOUCFG_
  SSHDCFG=$SSHDCFG_
  UDFILE=/var/lib/cloud/instance/user-data.txt
  get_user_data()
  {
    if test -r $UDFILE; then
      echo "Using cached user_data from cloud-init ..." 1>&2
      cat $UDFILE
    else
      echo "Getting user_data from MDS ..." 1>&2
      #otc.sh mds user_data 2>/dev/null
      curl -m6 -sS http://169.254.169.254/openstack/latest/user_data 2>/dev/null || \
      curl -sS http://169.254.169.254/openstack/latest/user_data 2>/dev/null
    fi
  }
fi

SHASFW2CFG="$(sha512sum $SFW2CFG)"
SHASFW2CUSTCFG="$(sha512sum $SFW2CUSTCFG)"

cond_sfw2_restart()
{
  # Check whether anything changed, otherwise skip and save two seconds
  NEWSHA="$(sha512sum $SFW2CFG)"
  NEWSHACUST="$(sha512sum $SFW2CUSTCFG)"
  if test "$SHASFW2CFG" != "$NEWSHA" -o "$SHASFW2CUSTCFG" != "$NEWSHACUST"; then
    echo "Restart SuSEfirewall2 ..."
    systemctl restart SuSEfirewall2
  else
    return 0
  fi
}

get_net()
{
  NET=$(echo "$USERDATA" | shyaml get-value otc.$1 | sed 's/- //')
  NET=$(echo $NET)
}

is_snat()
{
  echo "$1" | shyaml get-value otc.snat >/dev/null 2>&1
}

setup_sfw2()
{
  verbprint "Enable routing and masquerading"
  sed -i 's/^FW_ROUTE=.*$/FW_ROUTE="yes"/' $SFW2CFG
  sed -i 's/^FW_MASQUERADE=.*$/FW_MASQUERADE="yes"/' $SFW2CFG
  if test "$NEEDINT" = "1"; then
    if grep "^FW_DEV_INT=\"\"" $SFW2CFG >/dev/null 2>&1; then
      verbprint ".. setting fake int0 net"
      sed -i 's/^FW_DEV_INT=""/FW_DEV_INT="int0"/' $SFW2CFG
    fi
    sed -i 's@# ALLOWINT # \(iptables .*$\)@\1 # ALLOWINT #@' $SFW2CUSTCFG
    get_net internalnet
    ICMD2=""; ICMD3=""
    if test -n "$NET"; then 
      for net in $NET; do
        IPT2=`grep ALLOWINT2 $SFW2CUSTCFG | sed -e 's@^ *# ALLOWINT2 # @@'`
        ICMD2=$(echo "$IPT2; " | sed "s@INTERNALNET@$net@g")
        IPT3=`grep ALLOWINT3 $SFW2CUSTCFG | sed -e 's@^ *# ALLOWINT3 # @@'`
        ICMD3=$(echo "$IPT3; " | sed "s@INTERNALNET@$net@g")
      done
      echo -e "$ICMD2\n$ICMD3"
      sed -i "s@# ALLOWINT2 # \(iptables .*$\)@$ICMD2 # ALLOWINT2 #@" $SFW2CUSTCFG
      sed -i "s@# ALLOWINT3 # \(iptables .*$\)@$ICMD3 # ALLOWINT3 #@" $SFW2CUSTCFG
    fi
  else
    sed -i 's@\(iptables [^#]*\)# ALLOWINT #@# ALLOWINT # \1@' $SFW2CUSTCFG
    sed -i 's@\(iptables [^#;]*\).$# ALLOWINT2 #@# ALLOWINT2 # \1@' $SFW2CUSTCFG
    sed -i 's@\(iptables [^#;]*\).$# ALLOWINT3 #@# ALLOWINT3 # \1@' $SFW2CUSTCFG
  fi
  if test "$NEEDDMZ" = "1"; then
    if grep "^FW_DEV_DMZ=\"\"" $SFW2CFG >/dev/null 2>&1; then
      verbprint ".. setting fake dmz0 net"
      sed -i 's/^FW_DEV_DMZ=""/FW_DEV_DMZ="dmz0"/' $SFW2CFG
    fi
    sed -i 's@# ALLOWDMZ # \(iptables .*$\)@\1 # ALLOWDMZ #@' $SFW2CUSTCFG
    get_net dmznet
    ICMD2=""
    if test -n "$NET"; then 
      for net in $NET; do
        IPT2=`grep ALLOWDMZ2 $SFW2CUSTCFG | sed -e 's@^ *# ALLOWDMZ2 # @@'`
        ICMD2=$(echo "$IPT2; " | sed "s@DMZNET@$net@g")
      done
      sed -i "s@# ALLOWDMZ2 # \(iptables .*$\)@$ICMD2 # ALLOWDMZ2 #@" $SFW2CUSTCFG
    fi
  else
    sed -i 's@\(iptables [^#]*\)# ALLOWDMZ #@# ALLOWDMZ # \1@' $SFW2CUSTCFG
    sed -i 's@\(iptables [^#;]*\).$# ALLOWDMZ2 #@# ALLOWDMZ2 # \1@' $SFW2CUSTCFG
  fi
}

setup_inclass()
{
  verbprint "Enabling in-class routing"
  sed -i 's/^FW_ALLOW_CLASS_ROUTING=.*$/FW_ALLOW_CLASS_ROUTING="yes"/' $SFW2CFG
}

cfg_sfw2_par()
{ 
  MASQ=""
  while read line; do
    MASQ="$MASQ ${line#- }"
  done < <(shyaml get-value otc.$1 "" 2>/dev/null)
  MASQ="${MASQ# }"
  test -n "$MASQ" && verbprint "Setting $2 to $MASQ"
  #echo sed -i "s@^$2=.*\$@$2=\"$MASQ\"@" $SFW2CFG
  sed -i "s@^$2=.*\$@$2=\"$MASQ\"@" $SFW2CFG
}

NOINCLASS=""
cfg_sfw2_masq()
{
  echo "$USERDATA" | cfg_sfw2_par snat.masqnet FW_MASQ_NETS
  echo "$USERDATA" | cfg_sfw2_par snat.nomasqnet FW_NOMASQ_NETS
  echo "$USERDATA" | cfg_sfw2_par snat.fwdmasq FW_FORWARD_MASQ
}

localnet()
{
  # TODO: Support multiple networks
  LOCALNET=$(ip route | grep '^[0-9.]*/[0-9]* dev' | sed 's@^\([0-9.]*/[0-9]*\) dev.*@\1@' | head -n1)
  #echo "$LOCALNET" 1>&2
  sed "s@LOCALNET@${LOCALNET}@g"
}

# truth
is_true()
{
  test "$1" = "true" -o "$1" = "True" -o "$1" = "yes" -o "$1" = "Yes"
}

# persistroute: true
is_persistroute()
{
  VAL=$(echo "$1" | shyaml get-value otc.persistroute 2>&1) || return 2
  istrue "$VAL"
}

persistroute()
{
  verbprint "Saving routes ..."
  /sbin/route | grep eth0 | awk '{ print $1 "\t" $2 "\t" $3 "\t" $8 }' > /etc/sysconfig/network/routes
}

# SFW2:
#    VAR1: content1
#    VAR2: content2
is_sfw2()
{
  echo "$1" | shyaml get-value otc.SFW2 >/dev/null 2>&1
}

sfw2vars()
{
  while read line; do
    VAR=${line%%:*}
    VAL=${line#*: }
    sed -i "s@^$VAR=.*\$@$VAR=\"$VAL\"@" $SFW2CFG
    verbprint "Setting $VAR=\"$VAL\""
  done < <(echo "$USERDATA" | shyaml get-value otc.SFW2)
}

# autoupdate:
#    frequency: daily|weekly|monthly
#    category: security|secrec

is_autoupdate()
{
  echo "$1" | shyaml get-value otc.autoupdate >/dev/null 2>&1
}

autoupdate()
{
  FREQ=$(echo "$USERDATA" | shyaml get-value otc.autoupdate.frequency weekly)
  CAT=$(echo "$USERDATA" | shyaml get-value otc.autoupdate.category "security recommended")
  verbprint "Setup $FREQ autoupdate for $CAT"
  sed -i 's/^AOU_ENABLE_CRONJOB=.*$/AOU_ENABLE_CRONJOB="true"/' $AOUCFG
  #sed -i 's/^AOU_AUTO_AGREE_WITH_LICENSES=.*$/AOU_AUTO_AGREE_WITH_LICENSES="true"/' $AOUCFG
  sed -i "s/^AOU_PATCH_CATEGORIES=.*\$/AOU_PATCH_CATEGORIES=\"$CAT\"/" $AOUCFG
  # Set symlink opensuse.org-online_update -> /usr/lib/YaST2/bin/online_update
  rm -f /etc/cron.*ly/otc.t-systems.com-online_update
  ln -s /usr/lib/YaST2/bin/online_update /etc/cron.$FREQ/otc.t-systems.com-online_update
}

# movessh: PORT
is_movessh()
{
  echo "$1" | shyaml get-value otc.movessh >/dev/null 2>&1
}

# Implementation of movessh
# - Tweak sshd_config
# - Tweak SuSEfirewall2
# - Tweak fail2ban functionatliy
movessh()
{
  #global CHG
  OLDPORT=$(grep '^Port' $SSHDCFG 2>/dev/null | sed 's/^Port *\([0-9]*\).*$/\1/')
  if test -z "$OLDPORT"; then OLDP=22; else OLDP=$OLDPORT; fi
  NEWP=$(echo "$USERDATA" | shyaml get-value otc.movessh 22)
  #echo "Changing sshd from $OLDPORT ($OLDP) to $NEWP"
  if test "$OLDP" != "$NEWP"; then
    verbprint "Moving sshd from port $OLDP to $NEWP"
    # sshd_config
    if test -n "$OLDPORT"; then
      sed -i "s/^Port .*\$/Port $NEWP/" $SSHDCFG
    else
      echo "Port $NEWP" >> $SSHDCFG
    fi
    # SuSEfirewall2
    sed -i "/^FW_SERVICES/s@\([ \"]\)$OLDP\([ \"]\)@\1$NEWP\2@" $SFW2CFG
    # fail2ban
    sed -i "s@^SSHPORT=.*\$@SSHPORT=$NEWP@" $SFW2CUSTCFG
    CHG=4
    if test -z "$DEBUG"; then
      systemctl restart sshd
    fi
  fi
}

enable_custrules()
{
  verbprint "Use SFW2-snathelper"
  sed -i 's@^FW_CUSTOMRULES=.*@FW_CUSTOMRULES="/etc/sysconfig/scripts/SuSEfirewall2-snathelper"@' $SFW2CFG
}

FUNC="fw_custom_after_chain_creation() {"
BEGM="    #BEGIN SETUP_SNAT CATEGORIZE"
ENDM="    #END   SETUP_SNAT CATEGORIZE"

repl_custom()
{
    grep "^$BEGM" $SFW2CUSTCFG >/dev/null 2>&1
    NOINS=$?
    INRMV=0
    OLDIFS="$IFS"
    IFS=""
    while read line; do
      if test "$NOINS" = "1" -a "$line" = "$FUNC"; then
        echo "$line"
	echo "$BEGM"
        echo -e "$1"
        echo "$ENDM"
      elif test "$line" = "$BEGM"; then 
        INRMV=1; echo "$line"
	echo -e "$1"
      elif test "$line" = "$ENDM"; then
        INRMV=0; echo "$line"
      else
        if test "$INRMV" = "0"; then echo "$line"; fi
      fi
    done <$SFW2CUSTCFG >$SFW2CUSTCFG.new
    mv $SFW2CUSTCFG.new $SFW2CUSTCFG
    IFS="$OLDIFS"
}

# $1 net (int or dmz)
# $2 the *net
net_rules()
{
  RULES=""
  while read net; do
    net=${net#- }
    DEV=${net##*,}
    NET=${net%,*}
    if test "$NET" != "$net"; then DEV="-i $DEV"; else unset DEV; fi
    RULES="${RULES}    iptables -A INPUT   -j input_$1   $DEV -s $NET\n"
    RULES="${RULES}    iptables -A FORWARD -j forward_$1 $DEV -s $NET\n"
  done < <(echo "$USERDATA" | shyaml get-value otc.${2}net)
  echo "$RULES"
  if test -n "$RULES"; then
    verbprint "Redirecting traffic from $(echo "$USERDATA" | shyaml get-value otc.${2}net) to $1"
  fi
}


# $1 CAPITALNETNAME
# $2 lowercase
repl_net() 
{
  get_net $2
  USERDATA=$(echo "$USERDATA" | sed "s@$1@$NET@g")
  #echo "Replace $1 with $NET" 1>&2
}

net_categ()
{
  unset RUL
  if echo "$USERDATA" | shyaml get-value otc.internalnet >/dev/null 2>&1; then
    RUL=$(net_rules int internal)
    repl_net INTERNALNET internalnet
    NEEDINT=1
  fi
  if echo "$USERDATA" | shyaml get-value otc.dmznet >/dev/null 2>&1; then
    RUL="$RUL$(net_rules dmz dmz)"
    repl_net DMZNET dmznet
    NEEDDMZ=1
  fi
  if echo "$USERDATA" | shyaml get-value otc.externalnet >/dev/null 2>&1; then
    RUL="$RUL$(net_rules ext external)"
    repl_net EXTERNALNET externalnet
  fi
  if test -n "$RUL"; then
    repl_custom "$RUL"
    CHG=5
  fi
}

is_fail2ban()
{
  VAL=$(echo "$1" | shyaml get-value otc.fail2ban 2>&1) || return 2
  is_true $VAL
}

enable_fail2ban()
{
  verbprint "Enabling fail2ban"
  sed -i 's/^ENABLEF2B=.*$/ENABLEF2B="true"/' $SFW2CUSTCFG
}

disable_fail2ban()
{
  #verbprint "Disabling fail2ban"
  sed -i 's/^ENABLEF2B=.*$/ENABLEF2B="false"/' $SFW2CUSTCFG
}

is_custcateg()
{
  echo "$1" | shyaml get-value otc.internalnet >/dev/null 2>&1 && return 0
  echo "$1" | shyaml get-value otc.dmznet >/dev/null 2>&1 && return 0
  echo "$1" | shyaml get-value otc.externalnet >/dev/null 2>&1 && return 0
  return 1
}

is_addip()
{
  echo "$1" | shyaml get-value otc.addip >/dev/null 2>&1
}

SNAT_ADDIP=/opt/otc/snat/snat_addip.sh

addip()
{
  while read line; do
    DEV=${line%%: *}
    NET=${line#*: }
    for net in $NET; do
      if test -x $SNAT_ADDIP; then
        verbprint $SNAT_ADDIP $DEV $NET
        $SNAT_ADDIP $DEV $NET &
      else
        verbprint ip addr add $NET dev $DEV
        ip addr add $NET dev $DEV
      fi
    done
  done < <(echo "$USERDATA" | shyaml get-value otc.addip "")
}

is_addroute()
{
  echo "$1" | shyaml get-value otc.addroute >/dev/null 2>&1
}

addroute()
{
  while read line; do
    RT=${line#- }
    ip route add $RT
  done < <(echo "$USERDATA" | shyaml get-value otc.addroute "")
}


## MAIN
unset CHG
export USERDATA=$(get_user_data | localnet)
echo "$USERDATA" > /var/log/snat_setup_userdata
#USERDATA=$(echo "$USERDATA" | sed -e 's@^\([^-]*-\) PRIVATENET@\1 192.168/16\n\1 172.16/12\n\1 10/8@g' -e 's@PROVIDERNET@100.64/10@g')
USERDATA=$(echo "$USERDATA" | sed -e 's@PRIVATENET@192.168/16 172.16/12 10/8@g' -e 's@PROVIDERNET@100.64/10@g')
if is_addip "$USERDATA"; then addip; fi
if is_fail2ban "$USERDATA"; then enable_fail2ban; CHG=1; else disable_fail2ban; fi
# FIXME: Let's just look at the beginning ... 
if is_snat "$USERDATA"; then
  MASQ=$(echo "$USERDATA" | shyaml get-value otc.snat.masqnet "")
fi
if is_snat "$USERDATA" || is_fail2ban "$USERDATA" || is_custcateg "$USERDATA"; then
  enable_custrules
  net_categ
  CHG=1
fi
if is_snat "$USERDATA"; then
  if test "${MASQ:0:14}" != "- INTERNALNET" -a "${MASQ:0:8}" != "- DMZNET"; then
    setup_inclass
  fi
  setup_sfw2
  cfg_sfw2_masq
  CHG=2
fi
if is_sfw2 "$USERDATA"; then sfw2vars; CHG=3; fi
if is_persistroute "$USERDATA"; then persistroute; fi
if is_autoupdate "$USERDATA"; then autoupdate; fi
if is_movessh "$USERDATA"; then movessh; fi
if is_addroute "$USERDATA"; then addroute; fi

if test -n "$DEBUG"; then
  diff -u1 $SFW2CFG_ $SFW2CFG
  diff -u1 $SFW2CUSTCFG_ $SFW2CUSTCFG
  diff -u1 $AOUCFG_ $AOUCFG
  sudo diff -u1 $SSHDCFG_ $SSHDCFG
elif test -n "$CHG"; then
  cond_sfw2_restart
else
  true
fi

